# Simulations for AFT 2021 submission

# Fix the Type3 font error from the camera-ready submission form.
import matplotlib
matplotlib.rcParams["pdf.fonttype"] = 42
matplotlib.rcParams["ps.fonttype"] = 42

import collections
import itertools

from IPython.display import Javascript
import matplotlib.pyplot as plt
import numpy as np
import scipy.stats

display(Javascript("google.colab.output.setIframeHeight(0, true, { maxHeight: 9999 })"))

Transaction = collections.namedtuple("Transaction", ["price"])
Block = collections.namedtuple("Block", ["reserve_price", "transactions"])

num_slots = 100

def make_constant_demand(num_users):
  return lambda round: num_users

def make_step_demand(num_users_first, num_users_second):
  return lambda round: num_users_second if 200 <= round < 400 else num_users_first

def make_strawman_reserve_price_updater(alpha):
  assert 0 <= alpha <= 1

  def update_reserve_price(block):
    target_reserve_price = (
        sum(transaction.price for transaction in block.transactions) / num_slots)
    return alpha * target_reserve_price + (1 - alpha) * block.reserve_price

  return update_reserve_price

def make_welfare_reserve_price_updater(alpha, delta):
  assert 0 <= alpha <= 1
  assert 0 <= delta < float("inf")

  def update_reserve_price(block):
    if len(block.transactions) < num_slots:
      target_reserve_price = (
          sum(
              min(transaction.price, (1 + delta) * block.reserve_price)
              for transaction in block.transactions
          ) / num_slots)
    else:
      target_reserve_price = (1 + delta) * block.reserve_price
    return alpha * target_reserve_price + (1 - alpha) * block.reserve_price

  return update_reserve_price

def make_eip_1559_reserve_price_updater(alpha, delta):
  # This matches the EIP-1559 specification when alpha = 1/8 and delta = 1.

  assert 0 <= alpha <= 1
  assert 0 <= delta < float("inf")

  def update_reserve_price(block):
    target_reserve_price = (
        (len(block.transactions) / num_slots) * (1 + delta) * block.reserve_price)
    return alpha * target_reserve_price + (1 - alpha) * block.reserve_price

  return update_reserve_price

def run_simulation(
    sample_prices,
    update_reserve_price,
    num_rounds=600,
    get_num_users_in_round=make_constant_demand(1000),
    initial_reserve_price=1,
    randomized_allocation=False,
):
  reserve_price = initial_reserve_price
  blocks = []
  welfare_ratios = []

  for round in range(num_rounds):
    transactions = []
    for price in sample_prices(size=int(get_num_users_in_round(round))):
      transactions.append(Transaction(price=price))

    sorted_transactions = sorted(
        transactions, key=lambda transaction: transaction.price, reverse=True)
    optimal_block_transactions = sorted_transactions[:num_slots]
    optimal_block_welfare = sum(
        transaction.price for transaction in optimal_block_transactions)

    eligible_transactions = [
        transaction for transaction in transactions
        if transaction.price >= reserve_price]

    if not randomized_allocation:
      eligible_transactions.sort(
          key=lambda transaction: transaction.price, reverse=True)

    block_transactions = eligible_transactions[:num_slots]
    block_welfare = sum(
        transaction.price for transaction in block_transactions)

    block = Block(reserve_price=reserve_price, transactions=block_transactions)
    blocks.append(block)

    assert len(block.transactions) <= num_slots
    assert all(transaction.price >= reserve_price for transaction in block.transactions)

    welfare_ratio = block_welfare / optimal_block_welfare
    welfare_ratios.append(welfare_ratio)

    reserve_price = update_reserve_price(block)

  return blocks, welfare_ratios

demands = [
    ("Constant(200)", make_constant_demand(200)),
    ("Constant(67)", make_constant_demand(67)),
    ("Step(200, 600)", make_step_demand(200, 600)),
]

distributions = [
    ("Constant(100)", scipy.stats.rv_discrete(values=([100], [1.0]))),
    ("Uniform(0, 200)", scipy.stats.uniform(loc=0, scale=200)),
    ("Exponential(100)", scipy.stats.expon(scale=100)),
    ("70 Pareto(1) - 70", scipy.stats.pareto(b=1, scale=70, loc=-70)),
]

reserve_price_updaters = [
    (["Strawman", "WDPP"][1], make_strawman_reserve_price_updater(1 / 16), False),
    (["Welfare-based", "TWDPP"][1], make_welfare_reserve_price_updater(1 / 16, 1), True),
    (["EIP-1559", "UDPP"][1], make_eip_1559_reserve_price_updater(1 / 16, 1), True),
]

for (
    (demand_label, get_num_users_in_round),
    (distribution_label, distribution),
) in itertools.product(
    demands,
    distributions,
):
  num_plots = 3
  figure, axes = plt.subplots(nrows=1, ncols=num_plots, figsize=(7 * num_plots, 5))

  num_rounds = 600
  num_users_in_rounds = [
      int(get_num_users_in_round(round)) for round in range(num_rounds)]
  axes[0].plot(num_users_in_rounds)

  for reserve_price_updater_label, update_reserve_price, randomized_allocation in reserve_price_updaters:
    sample_prices = distribution.rvs
    np.random.seed(1)
    blocks, welfare_ratios = run_simulation(
        sample_prices,
        update_reserve_price,
        num_rounds=num_rounds,
        get_num_users_in_round=get_num_users_in_round,
        initial_reserve_price=1,
        randomized_allocation=randomized_allocation,
    )

    reserve_prices = [block.reserve_price for block in blocks]
    axes[1].plot(reserve_prices, label=reserve_price_updater_label)

    #filled_slot_ratios = [len(block.transactions) / num_slots for block in blocks]
    #axes[2].plot(filled_slot_ratios, label=reserve_price_updater_label)

    axes[2].plot(welfare_ratios, label=reserve_price_updater_label)

  fontsize = 16

  figure.suptitle(
      "Demand curve: {} and Value distribution: {}".format(
          demand_label, distribution_label),
      fontsize=fontsize)

  for i, ax in enumerate(axes):
    ax.set_xlabel("Round", fontsize=fontsize)
    ax.grid(which="major", axis="y")
    if i != 0:
      ax.legend(prop={"size": fontsize})

  axes[0].set_ylabel("Number of bidders", fontsize=fontsize)
  axes[1].set_ylabel("Posted price", fontsize=fontsize)
  axes[2].set_ylabel("Achieved welfare / optimal welfare", fontsize=fontsize)
  axes[2].set_ylim([-0.05, 1.05])

  #plt.show()

  filename = (
      "demand-{}-distribution-{}.pdf"
      .format(demand_label, distribution_label)
      .replace(" ", "").replace("(", "").replace(")", ""))
  print("Saving {}...".format(filename))
  plt.savefig(filename, bbox_inches="tight")
